# Copyright 2018 Lenovo (Beijing) Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as object_base

from cyborg.objects import base
from cyborg.objects.deployable import Deployable
from cyborg.objects.driver_objects.driver_attach_handle import \
    DriverAttachHandle
from cyborg.objects.driver_objects.driver_attribute import DriverAttribute
from cyborg.objects import fields as object_fields


@base.CyborgObjectRegistry.register
class DriverDeployable(base.DriverObjectBase,
                       object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'name': object_fields.StringField(nullable=False),
        'num_accelerators': object_fields.IntegerField(nullable=False),
        'attribute_list': object_fields.ListOfObjectsField(
            'DriverAttribute', default=[], nullable=True),
        # TODO() add field related to local_memory or just store in the
        # attribute list?
        'attach_handle_list': object_fields.ListOfObjectsField(
            'DriverAttachHandle', default=[], nullable=True),
        'driver_name': object_fields.StringField(nullable=True)
    }

    def create(self, context, device_id, cpid_id):
        """Create a driver-side Deployable object into DB. This object will be
        stored in separate db tables: deployable & attach_handle &
        attribute table.
        """

        # first store in deployable table through Deployable Object.
        deployable_obj = Deployable(context=context,
                                    name=self.name,
                                    num_accelerators=self.num_accelerators,
                                    device_id=device_id,
                                    driver_name=self.driver_name
                                    )
        deployable_obj.create(context)
        # create attribute_list for this deployable
        if hasattr(self, 'attribute_list'):
            for driver_attr in self.attribute_list:
                driver_attr.create(context, deployable_obj.id)

        # create attach_handle_list for this deployable
        if hasattr(self, 'attach_handle_list'):
            for driver_attach_handle in self.attach_handle_list:
                driver_attach_handle.create(context, deployable_obj.id,
                                            cpid_id)

    def destroy(self, context, device_id):
        """delete one driver-side deployable by calling existing Deployable
        and AttachHandle Object. Use name&host to identify Deployable and
        attach_info to identify the AttachHandle
        """

        # get deployable_id by name, get only one value.
        dep_obj = Deployable.get_by_name_deviceid(context, self.name,
                                                  device_id)
        # delete attach_handle
        if hasattr(self, 'attach_handle_list'):
            for driver_ah_obj in self.attach_handle_list:
                # get attach_handle_obj, exist and only one.
                driver_ah_obj.destroy(context, dep_obj.id)
        # delete attribute_list
        if hasattr(self, 'attribute_list'):
            DriverAttribute.destroy(context, dep_obj.id)
        # delete dep_obj
        if dep_obj is not None:
            dep_obj.destroy(context)

    @classmethod
    def list(cls, context, device_id):
        """Form driver-side Deployable object list from DB for one device."""
        # get deployable_obj_list for one device_id
        dep_obj_list = Deployable.get_list_by_device_id(context, device_id)
        driver_dep_obj_list = []
        for dep_obj in dep_obj_list:
            # get driver_ah_obj_list for this dep_obj
            driver_ah_obj_list = DriverAttachHandle.list(context, dep_obj.id)
            # get driver_attr_obj_list fro this dep_obj
            driver_attr_obj_list = DriverAttribute.list(context, dep_obj.id)
            driver_dep_obj = cls(context=context,
                                 name=dep_obj.name,
                                 num_accelerators=dep_obj.num_accelerators,
                                 attribute_list=driver_attr_obj_list,
                                 attach_handle_list=driver_ah_obj_list)
            driver_dep_obj_list.append(driver_dep_obj)
        return driver_dep_obj_list

    @classmethod
    def get_by_name(cls, context, name):
        """Form driver-side Deployable object list from DB for one device."""
        # get deployable_obj_list for one device_id
        dep_obj = Deployable.get_by_name(context, name)
        driver_ah_obj_list = DriverAttachHandle.list(context, dep_obj.id)
        # get driver_attr_obj_list fro this dep_obj
        driver_attr_obj_list = DriverAttribute.list(context, dep_obj.id)
        driver_dep_obj = cls(context=context, name=dep_obj.name,
                             num_accelerators=dep_obj.num_accelerators,
                             attribute_list=driver_attr_obj_list,
                             attach_handle_list=driver_ah_obj_list)
        return driver_dep_obj



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import tempfile

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_service import periodic_task
from oslo_utils import uuidutils

from cyborg.accelerator.drivers.fpga.base import FPGADriver
from cyborg.accelerator.drivers.gpu import utils as gpu_utils
from cyborg.agent.resource_tracker import ResourceTracker
from cyborg.agent.rpcapi import AgentAPI
from cyborg.common import exception
from cyborg.conductor import rpcapi as cond_api
from cyborg.conf import CONF
from cyborg.image.api import API as ImageAPI


LOG = logging.getLogger(__name__)


class AgentManager(periodic_task.PeriodicTasks):
    """Cyborg Agent manager main class.

    API version history:

    | 1.0 - Initial version.

    """

    RPC_API_VERSION = '1.0'
    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, topic, host=None):
        super(AgentManager, self).__init__(CONF)
        self.topic = topic
        self.host = host or CONF.host
        self.fpga_driver = FPGADriver()
        self.cond_api = cond_api.ConductorAPI()
        self.agent_api = AgentAPI()
        self.image_api = ImageAPI()
        self._rt = ResourceTracker(host, self.cond_api)

    def periodic_tasks(self, context, raise_on_error=False):
        return self.run_periodic_tasks(context, raise_on_error=raise_on_error)

    def fpga_program(self, context, controlpath_id,
                     bitstream_uuid, driver_name):
        bitstream_uuid = str(bitstream_uuid)
        if not uuidutils.is_uuid_like(bitstream_uuid):
            raise exception.InvalidUUID(uuid=bitstream_uuid)
        download_path = tempfile.NamedTemporaryFile(suffix=".gbs",
                                                    prefix=bitstream_uuid)
        self.image_api.download(context,
                                bitstream_uuid,
                                dest_path=download_path.name)
        try:
            driver = self.fpga_driver.create(driver_name)
            ret = driver.program(controlpath_id, download_path.name)
            LOG.info('Driver program() API returned %s', ret)
        finally:
            LOG.debug('Remove tmp bitstream file: %s', download_path.name)
            os.remove(download_path.name)
        return ret

    @periodic_task.periodic_task(run_immediately=True)
    def update_available_resource(self, context, startup=True):
        """Update all kinds of accelerator resources from their drivers."""
        self._rt.update_usage(context)

    def create_vgpu_mdev(self, context, pci_addr, asked_type, ah_uuid):
        LOG.debug('Instantiate a mediated device')
        gpu_utils.create_mdev_privileged(pci_addr, asked_type, ah_uuid)

    def remove_vgpu_mdev(self, context, pci_addr, asked_type, ah_uuid):
        LOG.debug('Remove a vgpu mdev')
        gpu_utils.remove_mdev_privileged(pci_addr, asked_type, ah_uuid)



# Copyright 2018 Lenovo (Beijing) Co.,LTD.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import base as object_base

from cyborg.objects.attach_handle import AttachHandle
from cyborg.objects import base
from cyborg.objects import fields as object_fields
LOG = logging.getLogger(__name__)


@base.CyborgObjectRegistry.register
class DriverAttachHandle(base.DriverObjectBase,
                         object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'attach_type': object_fields.StringField(nullable=False),
        # PCI BDF or mediated device ID...
        'attach_info': object_fields.StringField(nullable=False),
        # The status of attach_handle, is in use or not.
        'in_use': object_fields.BooleanField(nullable=False, default=False)
    }

    def create(self, context, deployable_id, cpid_id):
        """Create a driver-side AttachHandle object, call AttachHandle
        Object to store in DB.
        """
        attach_handle_obj = AttachHandle(context=context,
                                         deployable_id=deployable_id,
                                         cpid_id=cpid_id,
                                         attach_type=self.attach_type,
                                         attach_info=self.attach_info,
                                         in_use=self.in_use
                                         )
        attach_handle_obj.create(context)

    def destroy(self, context, deployable_id):
        ah_obj = AttachHandle.get_ah_by_depid_attachinfo(context,
                                                         deployable_id,
                                                         self.attach_info)
        if ah_obj is not None:
            ah_obj.destroy(context)

    @classmethod
    def list(cls, context, deployable_id):
        """Form a driver-side attach_handle list for one deployable."""
        ah_obj_list = AttachHandle.get_ah_list_by_deployable_id(
            context, deployable_id)
        driver_ah_obj_list = []
        for ah_obj in ah_obj_list:
            driver_ah_obj = cls(context=context,
                                attach_type=ah_obj.attach_type,
                                attach_info=ah_obj.attach_info,
                                in_use=ah_obj.in_use)
            driver_ah_obj_list.append(driver_ah_obj)
        return driver_ah_obj_list



# Copyright 2018 Intel, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
Cyborg Intel FPGA driver implementation.
"""

from oslo_concurrency import processutils
from oslo_log import log as logging

from cyborg.accelerator.drivers.fpga.base import FPGADriver
from cyborg.accelerator.drivers.fpga.intel import sysinfo
from cyborg.common import exception
import cyborg.privsep

LOG = logging.getLogger(__name__)


@cyborg.privsep.sys_admin_pctxt.entrypoint
def _fpga_program_privileged(cmd_args):
    # NOTE(Sundar): If we take cmd as parameter, this function can
    # be abused to run arbitrary commands in privileged mode. So
    # only cmd_args are passed in.
    # TODO(Sundar) Do not hardcode fpgaconf.
    # Use right tool based on bitstream type.
    # TODO(s_shogo): For now, we cannot choose the proper command.
    # Needs more discussion.
    cmd = ["/usr/bin/fpgaconf"]
    cmd.extend(cmd_args)
    # processutils will log the command line.
    return processutils.execute(*cmd)


class IntelFPGADriver(FPGADriver):
    """Class for Intel FPGA drivers.
       Vendor should implement their specific drivers in this class.
    """
    VENDOR = "intel"

    def __init__(self, *args, **kwargs):
        pass

    def discover(self):
        return sysinfo.fpga_tree()

    def program(self, controlpath_id, image_file_path):
        """Program the FPGA with the provided bitstream image.

           TODO(Sundar): Need to handle retries.

           :param: controlpath_id
               Controlpath_id OVO
           :param: image_file_path
               String with the file path
           :returns: True on success, False on failure
        """
        if controlpath_id['cpid_type'] != "PCI":
            raise exception.InvalidType(obj='controlpath_id',
                                        type=controlpath_id['cpid_type'],
                                        expected='PCI')
        cmd_args = []
        bdf_dict = controlpath_id['cpid_info']
        # fitting format to the OPAE command.
        bdf = ['0x' + s for s in map(lambda x: bdf_dict[x],
               ["bus", "device", "function"])]
        for i in zip(["--bus", "--device", "--function"], bdf):
            cmd_args.extend(i)
        cmd_args.append(image_file_path)

        try:
            # TODO(s_shogo): We should consider returning
            # a meaningful error message from the driver rather than
            # a boolean success/failure.
            _fpga_program_privileged(cmd_args)
            return True
        except Exception:
            # NOTE(Sundar): processutils.exec will log the error.
            # TODO(Sundar): If retryable error, try again.
            return False



# Copyright 2021 Inspur, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_serialization import jsonutils
from unittest import mock

from cyborg.accelerator.drivers.fpga.xilinx.driver import XilinxFPGADriver
from cyborg.tests import base

XILINX_FPGA_INFO = ["0000:3b:00.0 Processing accelerators [1200]: "
                    "Xilinx Corporation Device [10ee:5000]\n"
                    "0000:3b:00.1 Processing accelerators [1200]: "
                    "Xilinx Corporation Device [10ee:5001]"]


def fake_output(arg):
    if arg == ['lspci', '-nnn', '-D']:
        return XILINX_FPGA_INFO


class TestXilinxFPGADriver(base.TestCase):

    def setUp(self):
        super(TestXilinxFPGADriver, self).setUp()

    @mock.patch('cyborg.accelerator.drivers.fpga.'
                'xilinx.sysinfo.lspci_privileged')
    def test_discover(self, mock_devices_for_vendor):
        mock_devices_for_vendor.side_effect = fake_output
        self.set_defaults(host='fake-host', debug=True)
        fpga_list = XilinxFPGADriver().discover()
        self.assertEqual(1, len(fpga_list))
        attach_handle_list = [
            {'attach_type': 'PCI',
             'attach_info': '{"bus": "3b", '
                            '"device": "00", '
                            '"domain": "0000", '
                            '"function": "0"}',
             'in_use': False},
            {'attach_type': 'PCI',
             'attach_info': '{"bus": "3b", '
                            '"device": "00", '
                            '"domain": "0000", '
                            '"function": "1"}',
             'in_use': False}
        ]
        attribute_list = [
            {'key': 'rc', 'value': 'FPGA'},
            {'key': 'trait0', 'value': 'CUSTOM_FPGA_XILINX'},
            {'key': 'trait1', 'value': 'CUSTOM_FPGA_PRODUCT_ID_5000'},
        ]
        expected = {
            'vendor': '10ee',
            'type': 'FPGA',
            'std_board_info': {"controller": "Processing accelerators",
                               "product_id": "5000"},
            'vendor_board_info': {"vendor_info": "fpga_vb_info"},
            'deployable_list':
                [
                    {
                        'num_accelerators': 1,
                        'driver_name': 'XILINX',
                        'name': 'fake-host_0000:3b:00.0',
                        'attach_handle_list': attach_handle_list,
                        'attribute_list': attribute_list
                    },
                ],
            'controlpath_id': {'cpid_info': '{"bus": "3b", '
                                            '"device": "00", '
                                            '"domain": "0000", '
                                            '"function": "0"}',
                               'cpid_type': 'PCI'}
        }
        fpga_obj = fpga_list[0]
        fpga_dict = fpga_obj.as_dict()
        fpga_dep_list = fpga_dict['deployable_list']
        fpga_attach_handle_list = (
            fpga_dep_list[0].as_dict()['attach_handle_list'])
        fpga_attribute_list = fpga_dep_list[0].as_dict()['attribute_list']
        attri_obj_data = []
        [attri_obj_data.append(attr.as_dict()) for attr in fpga_attribute_list]
        attribute_actual_data = sorted(attri_obj_data, key=lambda i: i['key'])
        self.assertEqual(expected['vendor'], fpga_dict['vendor'])
        self.assertEqual(expected['controlpath_id'],
                         fpga_dict['controlpath_id'])
        self.assertEqual(expected['std_board_info'],
                         jsonutils.loads(fpga_dict['std_board_info']))
        self.assertEqual(expected['vendor_board_info'],
                         jsonutils.loads(fpga_dict['vendor_board_info']))
        self.assertEqual(expected['deployable_list'][0]['num_accelerators'],
                         fpga_dep_list[0].as_dict()['num_accelerators'])
        self.assertEqual(expected['deployable_list'][0]['name'],
                         fpga_dep_list[0].as_dict()['name'])
        self.assertEqual(expected['deployable_list'][0]['driver_name'],
                         fpga_dep_list[0].as_dict()['driver_name'])
        self.assertEqual(2, len(fpga_attach_handle_list))
        self.assertEqual(attach_handle_list[0],
                         fpga_attach_handle_list[0].as_dict())
        self.assertEqual(attach_handle_list[1],
                         fpga_attach_handle_list[1].as_dict())
        self.assertEqual(attribute_list, attribute_actual_data)

    @mock.patch('cyborg.accelerator.drivers.fpga.xilinx.driver.'
                '_fpga_program_privileged')
    def test_program(self, mock_prog):
        bdf = '0000:3b:00:0'
        expect_cmd_args = ['program', '--device', bdf, '--base',
                           '--image', '/path/image']

        xilinx_driver = XilinxFPGADriver()
        cpid_info = {"domain": "0000", "bus": "3b",
                     "device": "00", "function": "0"}
        cpid = {'cpid_type': 'PCI', 'cpid_info': cpid_info}

        # program PF
        mock_prog.return_value = bytes([0])
        xilinx_driver.program(cpid, "/path/image")
        mock_prog.assert_called_with(expect_cmd_args)



# Copyright 2020 Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pecan
import wsme
from wsme import types as wtypes

from oslo_serialization import jsonutils

from cyborg.agent.rpcapi import AgentAPI
from cyborg.api.controllers import base
from cyborg.api.controllers import link
from cyborg.api.controllers import types
from cyborg.api import expose
from cyborg.common import authorize_wsgi
from cyborg.common import exception as exc
from cyborg import objects


class Deployable(base.APIBase):
    """API representation of a deployable.

    This class enforces type checking and value constraints, and converts
    between the internal object model and the API representation of
    a deployable.
    """

    uuid = types.uuid
    """The UUID of the deployable"""

    parent_id = types.integer
    """The parent ID of the deployable"""

    root_id = types.integer
    """The root ID of the deployable"""

    name = wtypes.text
    """The name of the deployable"""

    num_accelerators = types.integer
    """The number of accelerators of the deployable"""

    device_id = types.integer
    """The device on which the deployable is located"""

    attributes_list = wtypes.text
    """The json list of attributes of the deployable"""

    rp_uuid = types.uuid
    """The uuid of resource provider which represents this deployable"""

    driver_name = wtypes.text
    """The driver name of this deployables"""

    bitstream_id = wtypes.text
    """The id of bitstream which has been program in this deployable"""

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing a self link"""

    def __init__(self, **kwargs):
        super(Deployable, self).__init__(**kwargs)
        self.fields = []
        for field in objects.Deployable.fields:
            self.fields.append(field)
            setattr(self, field, kwargs.get(field, wtypes.Unset))

    def convert_with_link(self, obj_dep):
        api_dep = Deployable(**obj_dep.as_dict())
        url = pecan.request.public_url
        api_dep.links = [
            link.Link.make_link('self', url, 'deployables', api_dep.uuid),
            link.Link.make_link('bookmark', url, 'deployables', api_dep.uuid,
                                bookmark=True)
            ]
        query = {"deployable_id": obj_dep.id}
        attr_get_list = objects.Attribute.get_by_filter(pecan.request.context,
                                                        query)
        attributes_list = []
        for exist_attr in attr_get_list:
            attributes_list.append({exist_attr.key: exist_attr.value})
        api_dep.attributes_list = jsonutils.dumps(attributes_list)
        return api_dep


class DeployableCollection(Deployable):
    """API representation of a collection of deployables."""

    deployables = [Deployable]
    """A list containing deployable objects"""

    def convert_with_links(self, obj_deps):
        collection = DeployableCollection()
        collection.deployables = [
            self.convert_with_link(obj_dep) for obj_dep in obj_deps]
        return collection


class DeployablePatchType(types.JsonPatchType):

    _api_base = Deployable

    @staticmethod
    def internal_attrs():
        defaults = types.JsonPatchType.internal_attrs()
        return defaults + ['/name', '/num_accelerators']


class DeployablesController(base.CyborgController,
                            DeployableCollection):
    """REST controller for Deployables."""

    _custom_actions = {'program': ['PATCH']}

    # TODO(s_shogo): We will update the policy of deployable APIs,
    # and using the new default policy rules in the W or later.
    @authorize_wsgi.authorize_wsgi("cyborg:deployable", "program", False)
    @expose.expose(Deployable, types.uuid, body=[DeployablePatchType])
    def program(self, uuid, program_info):
        """Program a new deployable(FPGA).

        :param uuid: The uuid of the target deployable.
        :param program_info: JSON string containing what to program.
        :raise: FPGAProgramError: If fpga program failed raise exception.
        :return: If fpga program success return deployable object.
        """

        image_uuid = program_info[0]['value'][0]['image_uuid']
        # TODO(s_shogo): In W or later version we plan to add schema check,
        # which will help checking input parameters' format.
        # So we can remove this validation in the future.
        try:
            types.UUIDType().validate(image_uuid)
        except Exception:
            raise

        obj_dep = objects.Deployable.get(pecan.request.context, uuid)
        obj_dev = objects.Device.get_by_device_id(
            pecan.request.context,
            obj_dep.device_id
        )
        hostname = obj_dev.hostname
        driver_name = obj_dep.driver_name
        cpid_list = obj_dep.get_cpid_list(pecan.request.context)
        controlpath_id = cpid_list[0]
        controlpath_id['cpid_info'] = jsonutils.loads(
            cpid_list[0]['cpid_info'])
        self.agent_rpcapi = AgentAPI()
        ret = self.agent_rpcapi.fpga_program(
            pecan.request.context,
            hostname,
            controlpath_id,
            image_uuid,
            driver_name,
            )
        if ret:
            return self.convert_with_link(obj_dep)
        else:
            raise exc.FPGAProgramError(ret=ret)

    @authorize_wsgi.authorize_wsgi("cyborg:deployable", "get_one")
    @expose.expose(Deployable, types.uuid)
    def get_one(self, uuid):
        """Retrieve information about the given deployable.

        :param uuid: UUID of a deployable.
        """
        obj_dep = objects.Deployable.get(pecan.request.context, uuid)
        return self.convert_with_link(obj_dep)

    @authorize_wsgi.authorize_wsgi("cyborg:deployable", "get_all")
    @expose.expose(DeployableCollection, wtypes.ArrayType(types.FilterType))
    def get_all(self, filters=None):
        """Retrieve a list of deployables.
        :param filters: a filter of FilterType to get deployables list by
        filter.
        """
        filters_dict = {}
        if filters:
            for filter in filters:
                filters_dict.update(filter.as_dict())
        context = pecan.request.context
        obj_deps = objects.Deployable.list(context, filters=filters_dict)
        return self.convert_with_links(obj_deps)



"""new_db_schema

Revision ID: ede4e3f1a232
Revises: d6f033d8fa5b
Create Date: 2018-11-27 22:00:52.080713

"""

# revision identifiers, used by Alembic.
revision = 'ede4e3f1a232'
down_revision = 'd6f033d8fa5b'

from alembic import op
import sqlalchemy as sa

# TODO: The enum value should be further discussed.
state = sa.Enum('Initial', 'Bound', 'BindFailed', name='state')
substate = sa.Enum('Initial', name='substate')
attach_type = sa.Enum('PCI', 'MDEV', name='attach_type')
cpid_type = sa.Enum('PCI', name='cpid_type')
control_type = sa.Enum('PCI', name='control_type')
device_type = sa.Enum('GPU', 'FPGA', name='device_type')


def upgrade():
    # drop old table: deployable, accelerator
    op.drop_table('attributes')
    op.drop_table('deployables')
    op.drop_table('accelerators')

    op.create_table(
        'devices',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        sa.Column('type', device_type, nullable=False),
        sa.Column('vendor', sa.String(length=255), nullable=False),
        sa.Column('model', sa.String(length=255), nullable=False),
        sa.Column('std_board_info', sa.Text(), nullable=True),
        sa.Column('vendor_board_info', sa.Text(), nullable=True),
        sa.Column('hostname', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'deployables',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        sa.Column('parent_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete='CASCADE'),
                  nullable=True),
        sa.Column('root_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete='CASCADE'),
                  nullable=True),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('num_accelerators', sa.Integer(), nullable=False),
        sa.Column('device_id', sa.Integer(),
                  sa.ForeignKey('devices.id', ondelete="RESTRICT"),
                  nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.Index('deployables_parent_id_idx', 'parent_id'),
        sa.Index('deployables_root_id_idx', 'root_id'),
        sa.Index('deployables_device_id_idx', 'device_id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'attributes',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        sa.Column('deployable_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete="RESTRICT"),
                  nullable=False, index=True),
        sa.Column('key', sa.Text(), nullable=False),
        sa.Column('value', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'controlpath_ids',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        sa.Column('device_id', sa.Integer(),
                  sa.ForeignKey('devices.id', ondelete="RESTRICT"),
                  nullable=False, index=True),
        sa.Column('cpid_type', cpid_type, nullable=False),
        sa.Column('cpid_info', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'attach_handles',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        sa.Column('deployable_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete="RESTRICT"),
                  nullable=False),
        sa.Column('cpid_id', sa.Integer(),
                  sa.ForeignKey('controlpath_ids.id', ondelete="RESTRICT"),
                  nullable=False),
        sa.Column('in_use', sa.Boolean(), default=False),
        sa.Column('attach_type', attach_type, nullable=False),
        sa.Column('attach_info', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.Index('attach_handles_deployable_id_idx', 'deployable_id'),
        sa.Index('attach_handles_cpid_id_idx', 'cpid_id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'device_profiles',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('profile_json', sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', 'name',
                            name='uniq_device_profiles0uuid0name'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )

    op.create_table(
        'extended_accelerator_requests',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('uuid', sa.String(length=36), nullable=False, unique=True),
        # NOTICE: we don't have project related constraints in Stein Release,
        # set nullable=True but keep this field for further expansion.
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('state', state, nullable=False, default='Initial'),
        sa.Column('device_profile_id', sa.Integer(),
                  sa.ForeignKey('device_profiles.id', ondelete="RESTRICT"),
                  nullable=False),
        sa.Column('hostname', sa.String(length=255), nullable=True),
        sa.Column('device_rp_uuid', sa.String(length=36), nullable=True),
        sa.Column('device_instance_uuid', sa.String(length=36),
                  nullable=True),
        sa.Column('attach_handle_id', sa.Integer(),
                  sa.ForeignKey('attach_handles.id', ondelete="RESTRICT"),
                  nullable=True),
        # Cyborg Private Fields begin here.
        sa.Column('substate', substate, nullable=False, default='Initial'),
        sa.Column('deployable_id', sa.Integer(),
                  sa.ForeignKey('deployables.id', ondelete="RESTRICT"),
                  nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.Index('extArqs_project_id_idx', 'project_id'),
        sa.Index('extArqs_device_profile_id_idx', 'device_profile_id'),
        sa.Index('extArqs_device_rp_uuid_idx', 'device_rp_uuid'),
        sa.Index('extArqs_device_instance_uuid_idx', 'device_instance_uuid'),
        sa.Index('extArqs_attach_handle_id_idx', 'attach_handle_id'),
        sa.Index('extArqs_deployable_id_idx', 'deployable_id'),
        mysql_ENGINE='InnoDB',
        mysql_DEFAULT_CHARSET='UTF8'
    )
